#include "stdafx.h"

#include <imagehlp.h>
#include <mbstring.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <vector>
#include <algorithm>

#include "BeckyAPI.h"
#include "BkCommon.h"
#include "resource.h"
#include "Util.h"
#include "Logger.h"
#include "ProgressDlg.h"
#include "BkBsfilProfile.h"
#include "AddressInfo.h"
#include "DllMain.h"
#include "BkBsfilWrapper.h"

////////////////////////////////////////////////////////////////////////
// BkBsfilWrapper ̂߂̎
////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
// }N
////////////////////////////////////////////////////////////////////////

#define SPAM_HEADER_PREFIX	"X-"

#define GET_SPAM_HEADER_FIELD_NAME(refSpamHeaderFieldName_, refbkProfile_) do { \
	*(refSpamHeaderFieldName_) = SPAM_HEADER_PREFIX; \
	*(refSpamHeaderFieldName_) += (refbkProfile_)->GetSpamHeaderFieldName(); \
} while (false)

#define GET_SPAM_HEADER(refSpamHeader_, refbkProfile_) do { \
	GET_SPAM_HEADER_FIELD_NAME((refSpamHeader_), (refbkProfile_)); \
	*(refSpamHeader_) += ": "; \
	*(refSpamHeader_) += (refbkProfile_)->GetSpamHeaderFieldBody(); \
} while (false)

//----------------------------------------------------------------------
// Xbhւ̃p[^
//----------------------------------------------------------------------

typedef struct BsfilStdinThreadParam_ {
	const char* pInput;
	int inputLength;
	HANDLE* phPipeWrite;
} BsfilStdinThreadParam;

//----------------------------------------------------------------------
// bsfilter ̊{̃R}hC𐶐
//----------------------------------------------------------------------

static std::string BuildBsfilterBaseCommandLine()
{
	char command[8192];

	_snprintf(command, sizeof(command), "\"%s\" --homedir \"%s\" --config-file \"%s\"", g_bsfilter, g_dbFolder, g_bsfilterConfig);
	FORCE_ASCIZ(command);

	return command;
}

//----------------------------------------------------------------------
// bsfilter ̕W͂ɏރXbh
//----------------------------------------------------------------------

static DWORD WINAPI WriteBsfilStdinThread(LPVOID pParam)
{
#define __FUNCNAME__ "WriteBsfilStdinThread()"
	char* pcfnname = "";

	LOGGING_DEBUG("begin.");

	BsfilStdinThreadParam* pBsfilParam = (BsfilStdinThreadParam*) pParam;

	const char* pInput = pBsfilParam->pInput;
	int inputLength = pBsfilParam->inputLength;
	HANDLE* phPipeWrite = pBsfilParam->phPipeWrite;

	pcfnname = "Util::WriteToFile()";
	BOOL rc = Util::WriteToFile(*phPipeWrite, pInput, inputLength);
	if (!rc) {
		LOGGING_ERROR("%s.", pcfnname);
	}
	LOGGING_DEBUG("%s: succeeded.", pcfnname);

	CloseHandle(*phPipeWrite);
	*phPipeWrite = INVALID_HANDLE_VALUE;

	LOGGING_DEBUG("end.");

	return 0;

#undef __FUNCNAME__
}

//----------------------------------------------------------------------
// bsfilter ̐ݒt@C쐬
//----------------------------------------------------------------------

static void CreateBsfilterConfig()
{
#define __FUNCNAME__ "CreateBsfilterConfig()"
	char* pcfnname = "";

	LOGGING_DEBUG("begin.");

	HANDLE hFile = INVALID_HANDLE_VALUE;

	do {

		pcfnname = "CreateFile()";
		hFile = CreateFile(g_bsfilterConfig, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
		if (hFile == INVALID_HANDLE_VALUE) {
			LOGGING_ERROR("%s: LastError: [%lu], g_bsfilterConfig: [%s].", pcfnname, GetLastError(), g_bsfilterConfig);
			break;
		}
		LOGGING_DEBUG("%s: succeeded.", pcfnname);

		// 󂾂Ǝ₵̂ŁA^CgƂăRgo͂
		char* pTitle = "## bsfilter.conf\r\n";

		DWORD length = strlen(pTitle);
		pcfnname = "Util::WriteToFile()";
		BOOL rc = Util::WriteToFile(hFile, pTitle, length);
		if (!rc) {
			LOGGING_ERROR("%s.", pcfnname);
		}
		LOGGING_DEBUG("%s: succeeded.", pcfnname);

	} while (false);

	CloseHandle(hFile);

	LOGGING_DEBUG("end.");

#undef __FUNCNAME__
}

//----------------------------------------------------------------------
// [{bNXw̃[{bNXɕύX
//----------------------------------------------------------------------

static std::string ChangeMailboxId(std::string folderId, std::string mailBoxNew)
{
	// ΃tH_ʒuɕύXB
	// ) XXXXXXXX.mb\XXX\  YYYYYYYY.mb\XXX\ ɕύX(YYYYYYYY ͎w肳ꂽ[{bNX)
	const char *pFolderId = folderId.c_str();
	const char *pFolder = (char*) _mbschr((unsigned char*) pFolderId, '\\');
	if (pFolder && pFolderId < pFolder) {

		std::string folderName = &pFolder[1];

		if (!Util::EndsWithPathDelimiter(mailBoxNew)) {
			mailBoxNew += "\\";
		}

		folderId = mailBoxNew + folderName;
	}

	return folderId;
}

//----------------------------------------------------------------------
// I𒆂̃[̃[ ID ̈ꗗ擾
//----------------------------------------------------------------------

static void GetSelectedMailIdList(std::vector<std::string>* pMailIdList)
{
	char mailId[256];
	int itemPosition = -1;
	do {

		itemPosition = bka.GetNextMail(itemPosition, mailId, sizeof(mailId), TRUE);
		if (itemPosition != -1) {
			pMailIdList->push_back(mailId);
		}

	} while (itemPosition != -1);
}

//----------------------------------------------------------------------
// sapm \wb_폜
//----------------------------------------------------------------------

static bool RemoveSpamHeader(std::string mailId, char* pMail, int mailLength)
{
	bool result = false;

	if (g_bkProfile.GetSpamHeaderFieldName().length() == 0) {
		return false;
	}

	std::string spamHeaderFieldName;
	GET_SPAM_HEADER_FIELD_NAME(&spamHeaderFieldName, &g_bkProfile);

	int indexFieldNameStart = 0;
	int indexFieldNameEnd = 0;
	int indexBodyStart = 0;
	int indexBodyEnd = 0;

	std::vector<char> mailNew;
	bool bRemove = Util::RemoveSpamHeader(pMail, mailLength, &spamHeaderFieldName, &mailNew);
	if (bRemove) {
		bka.SetSource(mailId.c_str(), &mailNew[0]);
		result = true;
	}

	return result;
}

//----------------------------------------------------------------------
// spam ړ̃tH_𓾂
//----------------------------------------------------------------------

static std::string GetSpamFolderId()
{
	std::string folderId;
	switch (g_bkProfile.GetProcSpam()) {
	case BkBsfilProfile::GoTrash:
		{
			std::string mailBox = bka.GetCurrentMailBox();
			if (!Util::EndsWithPathDelimiter(mailBox)) {
				mailBox += "\\";
			}
			folderId = mailBox + "!Trash\\";
		}
		break;

	case BkBsfilProfile::GoFolder:
		folderId = g_bkProfile.GetSpamFolder();

		if (g_bkProfile.GetGoFolderSameRelativeFolderPath()) {
			// [{bNX݂̃[{bNXɕύX
			std::string mailBox = bka.GetCurrentMailBox();
			folderId = ChangeMailboxId(folderId, mailBox);
		}
		break;
	}

	return folderId;
}

//----------------------------------------------------------------------
// sapm \wb_ǉ
//----------------------------------------------------------------------

static bool AddSpamHeader(std::string spamMailId, char* pMail, int mailLength)
{
	bool result = false;

	if (g_bkProfile.GetSpamHeaderFieldName().length() == 0 || g_bkProfile.GetSpamHeaderFieldBody().length() == 0) {
		return false;
	}

	std::string spamHeaderFieldName;
	GET_SPAM_HEADER_FIELD_NAME(&spamHeaderFieldName, &g_bkProfile);

	std::string spamHeader;
	GET_SPAM_HEADER(&spamHeader, &g_bkProfile);
	spamHeader += "\r\n";

	std::vector<char> mailNew;
	bool bAdd = Util::AddSpamHeader(pMail, mailLength, &spamHeaderFieldName, &spamHeader, &mailNew);
	if (bAdd) {
		bka.SetSource(spamMailId.c_str(), &mailNew[0]);
		result = true;
	}

	return result;
}

//----------------------------------------------------------------------
// [уhCzCgXg珜
//----------------------------------------------------------------------

static void RemoveMailOrDomainFromWhiteList(char* pMail)
{
	char name[1024];
	char email[512];

	std::string from = Util::GetHeaderFieldBody(pMail, "From");
	GetDecodedNameAndAddr(name, sizeof(name), email, sizeof(email), from.c_str());

	std::string domain = Util::GetAtNext(email);

	if (strlen(email) != 0 && g_whiteList.IsExist(email)) {
		g_whiteList.Delete(email);
	}

	if (domain.length() != 0 && g_whiteList.IsExist(domain.c_str())) {
		g_whiteList.Delete(domain.c_str());
	}
}

//----------------------------------------------------------------------
// [wK
//----------------------------------------------------------------------

static int LearnMail(std::vector<std::string>* pMaillist, std::vector<std::string>* pProcessedMailIdList, const char* pCommand, CProgressDlg& progressDlg)
{
	int result = 0;

	for (std::vector<std::string>::iterator it = pMaillist->begin(); it != pMaillist->end(); ++it) {

		std::string mailId = *it;

		char* pMail = bka.GetSource(mailId.c_str());
		if (!pMail) {
			continue;
		}

		int mailLength = strlen(pMail);
		std::string output;
		int rc = ExecBsfilter(pMail, mailLength, &output, "%s", pCommand);
		if (0 == rc) {
			++result;

			if (!g_bkProfile.GetDontRemoveSpamHeader()
			 && (strcmp(pCommand, "--add-clean") == 0 || strcmp(pCommand, "--sub-spam") == 0)) {
				// 莞у[wK clean ȃ[ spam wb_
				RemoveSpamHeader(mailId, pMail, mailLength);
			}

			if (g_bkProfile.GetRemoveSpamFromWhiteList()
			 && (strcmp(pCommand, "--add-spam") == 0 || strcmp(pCommand, "--sub-clean") == 0)) {
				 // 莞у[wK spam ȃ[zCgXg珜
				RemoveMailOrDomainFromWhiteList(pMail);
			}
		}

		bka.Free(pMail);

		if (pProcessedMailIdList) {
			pProcessedMailIdList->push_back(mailId);
		}

		// i\s
		progressDlg.StepIt();

		if (progressDlg.IsCancel()) {
			break;
		}
	}

	return result;
}

//----------------------------------------------------------------------
// [wKmf[^x[XXV
//----------------------------------------------------------------------

static void LearnAndUpdate(HWND hWnd, char* pCommand)
{
	int rc;
	std::string message;
	std::string title;

	message = CResourceManager::FormatResourceString(IDS_MSG_BEFORE_LEARN_EMAIL_AND_UPDATE_DATABASE,
			pCommand
		);
	title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_BEFORE_LEARN_EMAIL_AND_UPDATE_DATABASE,
			pCommand
		);

	if (!g_bkProfile.GetDisableBeforeMsgbox()) {
		rc = MessageBox(hWnd, message.c_str(), title.c_str(), MB_OKCANCEL | MB_ICONINFORMATION);
		if (rc != IDOK) {
			return;
		}
	}

	//
	// I𒆂̃[wKAmf[^x[XXV
	//
	std::vector<std::string> mailIdList;

	GetSelectedMailIdList(&mailIdList);

	// i\p
	HWND hWndMain = 0;
	HWND hWndDummy = 0;
	bka.GetWindowHandles(&hWndMain, &hWndDummy, &hWndDummy, &hWndDummy);

	CProgressDlg progressDlg(mailIdList.size(), title.c_str());
	progressDlg.Show(hWndMain);

	int learnCount = LearnMail(&mailIdList, 0, pCommand, progressDlg);

	std::string terminateStepMsg = CResourceManager::FormatResourceString(IDS_MSG_PROGRESS_TERMINATE);
	progressDlg.TerminateStep(terminateStepMsg.c_str());

	int updateStatus = 0;
	if (0 < learnCount) {
		std::string output;

		updateStatus = ExecBsfilter(0, 0, &output, "--update");
	}

	progressDlg.Hide();

	if (learnCount == 0 || (0 < learnCount && updateStatus == 0)) {

		// XV̏Ԃ𓾂
		std::string stsoutput;
		int status = ExecBsfilter(0, 0, &stsoutput, "--show-db-status");

		message = CResourceManager::FormatResourceString(IDS_MSG_LEARNED_EMAIL,
				learnCount, pCommand
			);
		if (status == 0) {
			Util::SameCrLf(stsoutput);
			message += "\r\n\r\n";
			message += stsoutput;
		}

		if (!g_bkProfile.GetDisableResultMsgbox()) {
			title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_EMAIL_LEARNING_RESULTS);

			MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONINFORMATION);
		}
	} else {
		title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_EMAIL_LEARNING_RESULTS);
		message = CResourceManager::FormatResourceString(IDS_MSG_EMAIL_FAILED_TO_LEARN);

		MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONERROR);
	}
}

//----------------------------------------------------------------------
// Oƃ[AhX𓾂
//----------------------------------------------------------------------

void GetDecodedNameAndAddr(char* name, int lenName, char* email, int lenEmail, const char* pSrc)
{
	GetNameAndAddr(name, lenName, email, lenEmail, pSrc);
	FORCE_ASCIZ_LEN(name, lenName);
	FORCE_ASCIZ_LEN(email, lenEmail);

	char* pdecodedname = 0;

	if (strlen(name) != 0) {
		char charset[128] = "";

		pdecodedname = bka.MIMEHeader(name, charset, sizeof(charset), FALSE);
		if (pdecodedname && strlen(pdecodedname) < (size_t) lenName) {
			strcpy(name, pdecodedname);
		}
	}

	if (pdecodedname) {
		bka.Free(pdecodedname);
	}
}

//----------------------------------------------------------------------
// wb_tB[hɑΉwb_tB[h{fBfR[hĕԂ
//----------------------------------------------------------------------

std::string GetDecodedHeaderFieldBody(const char* pSrc, const char* pHeaderFieldName)
{
	std::string headerFieldBody = Util::GetHeaderFieldBody(pSrc, pHeaderFieldName);

	if (headerFieldBody.length() != 0) {
		char charset[128];

		char* pdecodedFieldBody = bka.MIMEHeader(headerFieldBody.c_str(), charset, sizeof(charset), FALSE);
		if (pdecodedFieldBody && 0 < strlen(pdecodedFieldBody)) {
			headerFieldBody = pdecodedFieldBody;
		}

		if (pdecodedFieldBody) {
			bka.Free(pdecodedFieldBody);
		}
	}

	return headerFieldBody;
}

//----------------------------------------------------------------------
// fR[hԂ
//----------------------------------------------------------------------

std::string GetDecodedString(std::string& str)
{
	std::string result = str;

	if (result.length() != 0) {
		char charset[128];

		char* pdecodedResult = bka.MIMEHeader(result.c_str(), charset, sizeof(charset), FALSE);
		if (pdecodedResult && 0 < strlen(pdecodedResult)) {
			result = pdecodedResult;
		}

		if (pdecodedResult) {
			bka.Free(pdecodedResult);
		}
	}

	return result;
}

//----------------------------------------------------------------------
// M悩烁[AhẌꗗ𓾂
//----------------------------------------------------------------------

void AddSenderAddrToList(std::map<std::string, std::string>* pAddrList, std::string& to)
{
	std::vector<char> toArr(to.length() + 1);
	strcpy(&toArr[0], to.c_str());

	char* paddr = TokenAddr(&toArr[0]);
	while (paddr) {
		char name[1024];
		char email[512];

		GetDecodedNameAndAddr(name, sizeof(name), email, sizeof(email), paddr);
		if (strlen(email) != 0) {
			(*pAddrList)[email] = name;
		}

		paddr = TokenAddr(0);
	}
}

//----------------------------------------------------------------------
// bsfilter s
//----------------------------------------------------------------------

int ExecBsfilter(const char* pInput, int inputLength, std::string* pOutput, char* fmt, ...)
{
#define __FUNCNAME__ "ExecBsfilter()"
	char* pcfnname = "";

	CSLock cslock(&g_csDatabase);

	LOGGING_DEBUG("begin.");

	int result = -1;

	HANDLE hPipeWrite = INVALID_HANDLE_VALUE;
	HANDLE hPipeWriteTmp = INVALID_HANDLE_VALUE;
	HANDLE hPipeBsfilStdin = INVALID_HANDLE_VALUE;

	HANDLE hPipeRead = INVALID_HANDLE_VALUE;
	HANDLE hPipeReadTmp = INVALID_HANDLE_VALUE;
	HANDLE hPipeBsfilStdout = INVALID_HANDLE_VALUE;
	HANDLE hPipeBsfilStderr = INVALID_HANDLE_VALUE;

	PROCESS_INFORMATION pi;

	HANDLE hThread = 0;
	BsfilStdinThreadParam bsfilParam;

	va_list args;
	BOOL rc;

	va_start(args, fmt);

	ZeroMemory(&pi, sizeof(pi));

	do {

		SECURITY_ATTRIBUTES sa;
		STARTUPINFO si;

		if (pInput) {

			//
			// bsfilter ̕W͂ɏނ߂̃pCv쐬
			//

			ZeroMemory(&sa, sizeof(sa));
			sa.nLength = sizeof(sa);
			sa.bInheritHandle = TRUE;

			pcfnname = "CreatePipe()";
			rc = CreatePipe(&hPipeBsfilStdin, &hPipeWriteTmp, &sa, 0);
			if (!rc) {
				LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
				break;
			}
			LOGGING_DEBUG("%s: succeeded.", pcfnname);

			// eŎgݗpnh́ApȂ悤ɐݒ肵Ãnh͕
			pcfnname = "DuplicateHandle()";
			rc = DuplicateHandle(GetCurrentProcess(), hPipeWriteTmp, GetCurrentProcess(), &hPipeWrite, 0, FALSE, DUPLICATE_SAME_ACCESS);
			if (!rc) {
				LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
				break;
			}
			LOGGING_DEBUG("%s: succeeded.", pcfnname);

			CloseHandle(hPipeWriteTmp);
			hPipeWriteTmp = INVALID_HANDLE_VALUE;
		}

		if (pOutput) {

			//
			// bsfilter ̕Wo͂ǂݍނ߂̃pCv쐬
			//

			ZeroMemory(&sa, sizeof(sa));
			sa.nLength = sizeof(sa);
			sa.bInheritHandle = TRUE;

			pcfnname = "CreatePipe()";
			rc = CreatePipe(&hPipeReadTmp, &hPipeBsfilStdout, &sa, 0);
			if (!rc) {
				LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
				break;
			}
			LOGGING_DEBUG("%s: succeeded.", pcfnname);

			// bsfilter ̃G[o͗pɕ
			pcfnname = "DuplicateHandle()";
			rc = DuplicateHandle(GetCurrentProcess(), hPipeBsfilStdout, GetCurrentProcess(), &hPipeBsfilStderr, 0, TRUE, DUPLICATE_SAME_ACCESS);
			if (!rc) {
				LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
				break;
			}
			LOGGING_DEBUG("%s: succeeded.", pcfnname);

			// eŎgǂݍݗpnh́ApȂ悤ɐݒ肵Ãnh͕
			pcfnname = "DuplicateHandle()";
			rc = DuplicateHandle(GetCurrentProcess(), hPipeReadTmp, GetCurrentProcess(), &hPipeRead, 0, FALSE, DUPLICATE_SAME_ACCESS);
			if (!rc) {
				LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
				break;
			}
			LOGGING_DEBUG("%s: succeeded.", pcfnname);

			CloseHandle(hPipeReadTmp);
			hPipeReadTmp = INVALID_HANDLE_VALUE;
		}

		//
		// bsfilter N
		//

		ZeroMemory(&si, sizeof(si));
		si.cb = sizeof(si);
		if (pInput || pOutput) {
			si.dwFlags = STARTF_USESTDHANDLES;

			si.hStdInput = GetStdHandle(STD_INPUT_HANDLE);
			si.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE);
			si.hStdError = GetStdHandle(STD_ERROR_HANDLE);
		}

		if (pInput) {
			si.hStdInput = hPipeBsfilStdin;
		}
		if (pOutput) {
			si.hStdOutput = hPipeBsfilStdout;
			si.hStdError = hPipeBsfilStderr;
		}

		std::string bsfilbaseCommand = BuildBsfilterBaseCommandLine();
		bsfilbaseCommand += " ";
		bsfilbaseCommand += fmt;

		char command[8192];
		_vsnprintf(command, sizeof(command), bsfilbaseCommand.c_str(), args);
		FORCE_ASCIZ(command);

		pcfnname = "CreateProcess()";
		rc = CreateProcess(
						0, command,
						0, 0,
						TRUE,
						CREATE_DEFAULT_ERROR_MODE | CREATE_NEW_PROCESS_GROUP | CREATE_NO_WINDOW,
						0, 0,
						&si, &pi
					);

		if (!rc) {
			LOGGING_ERROR("%s: LastError: [%lu], command: [%s].", pcfnname, GetLastError(), command);
			break;
		}
		LOGGING_DEBUG("%s: succeeded. command: [%s].", pcfnname, command);

		if (pInput) {

			//
			// bsfilter ̕W͂ɏ
			//

			ZeroMemory(&bsfilParam, sizeof(bsfilParam));

			bsfilParam.pInput = pInput;
			bsfilParam.inputLength = inputLength;
			bsfilParam.phPipeWrite = &hPipeWrite;

			pcfnname = "CreateThread()";
			hThread = CreateThread(0, 0, WriteBsfilStdinThread, &bsfilParam, 0, 0);
			if (!hThread) {
				LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());

				TerminateProcess(pi.hProcess, -1);
				break;
			}
			LOGGING_DEBUG("%s: succeeded.", pcfnname);
		}

		DWORD waitStatus;

		if (pOutput) {

			//
			// bsfilter o͂e擾
			//

			do {

				pcfnname = "WaitForSingleObject()";
				waitStatus = WaitForSingleObject(pi.hProcess, 1000);
				if (waitStatus == WAIT_FAILED) {
					LOGGING_ERROR("%s: LastError: [%lu], waitStatus: [%lu].", pcfnname, GetLastError(), waitStatus);
					break;
				}
				LOGGING_DEBUG("%s: succeeded. waitStatus: [%lu].", pcfnname, waitStatus);

				DWORD readBytes;

				pcfnname = "PeekNamedPipe()";
				rc = PeekNamedPipe(hPipeRead, 0, 0, 0, &readBytes, 0);
				if (!rc) {
					LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
				} else {
					LOGGING_DEBUG("%s: succeeded. readBytes: [%lu].", pcfnname, readBytes);

					if (0 < readBytes) {

						char buffer[8192];

						pcfnname = "ReadFile()";
						if (ReadFile(hPipeRead, buffer, sizeof(buffer) - 1, &readBytes, 0)) {

							LOGGING_DEBUG("%s: succeeded. readBytes: [%lu].", pcfnname, readBytes);

							buffer[readBytes] = '\0';
							*pOutput += buffer;

						} else {
							LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
						}
					}
				}

			} while (waitStatus != WAIT_OBJECT_0);

			LOGGING_DEBUG(pOutput, "pOutput->length(): [%lu].", pOutput->length());
		}

		//
		// bsfilter ̕W͂ɏރXbh̏I҂
		//

		if (hThread) {
			pcfnname = "WaitForSingleObject()";
			waitStatus = WaitForSingleObject(hThread, 60000);
			if (waitStatus == WAIT_OBJECT_0) {
				LOGGING_DEBUG("%s: succeeded.", pcfnname);
			} else {
				LOGGING_ERROR("%s: LastError: [%lu], waitStatus: [%lu].", pcfnname, GetLastError(), waitStatus);
			}
		}

		//
		// bsfilter ̏I҂
		//

		DWORD exitCode;

		pcfnname = "WaitForSingleObject()";
		waitStatus = WaitForSingleObject(pi.hProcess, 60000);
		if (waitStatus == WAIT_OBJECT_0) {
			LOGGING_DEBUG("%s: succeeded.", pcfnname);

			pcfnname = "GetExitCodeProcess()";
			if (GetExitCodeProcess(pi.hProcess, &exitCode)) {
				result = exitCode;
				LOGGING_DEBUG("%s: succeeded. exitCode: [%lu].", pcfnname, exitCode);
			} else {
				LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
			}
		} else {
			LOGGING_ERROR("%s: LastError: [%lu], waitStatus: [%lu].", pcfnname, GetLastError(), waitStatus);
		}

	} while (false);

	//
	// n
	//

	va_end(args);

	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);

	CloseHandle(hThread);

	CloseHandle(hPipeWrite);
	CloseHandle(hPipeWriteTmp);
	CloseHandle(hPipeBsfilStdin);

	CloseHandle(hPipeRead);
	CloseHandle(hPipeReadTmp);
	CloseHandle(hPipeBsfilStdout);
	CloseHandle(hPipeBsfilStderr);

	LOGGING_DEBUG("end.");

	return result;

#undef __FUNCNAME__
}

//----------------------------------------------------------------------
// Becky! ̃o[W`FbNs
//----------------------------------------------------------------------

bool IsValidBeckyVersion(const char* bkaVersion)
{
	if (!bkaVersion) {
		return false;
	}

	// NOTE: Becky! ̃o[W̃tH[}bg "X.YY.ZZ" ŁA
	//       X: W[o[WAYY: }Ci[o[WAZZ: rWԍ
	//       ƂȂB
	std::vector<std::string> versionList;
	std::string csBkaVersion = bkaVersion;

	int pos = 0;
	int nextpos = 0;
	while ((nextpos = csBkaVersion.find_first_of('.', pos)) != csBkaVersion.npos) {
		versionList.push_back(csBkaVersion.substr(pos, nextpos - pos));
		pos = nextpos + 1;
	}
	if (0 <= pos) {
		versionList.push_back(csBkaVersion.substr(pos));
	}

	int majorVersion = 0;
	int minorVersion = 0;
	int revisionNumber = 0;
	if (1 <= versionList.size()) {
		majorVersion = atoi(versionList[0].c_str());
	}
	if (2 <= versionList.size()) {
		minorVersion = atoi(versionList[1].c_str());
	}
	if (3 <= versionList.size()) {
		revisionNumber = atoi(versionList[2].c_str());
	}

	if (majorVersion < 2) {
		return false;
	} else if (majorVersion == 2 && minorVersion < 50) {
		return false;
	}

	return true;
}

//----------------------------------------------------------------------
// s
//----------------------------------------------------------------------

bool Initialize()
{
#define __FUNCNAME__ "Initialize()"
	char* pcfnname = "";

	BOOL rc;

	// P[ݒ肷
	// NOTE: rh /ML ȊOɂꍇA^CL鑼 dll 
	//       AvP[V{̂ւ̉êȂׂ̂邱ƁB
	std::locale::global(std::locale("japanese"));

	// ݒ̔rpIuWFNg
	InitializeCriticalSection(&g_csSetting);

	// f[^x[X̔rpIuWFNg
	InitializeCriticalSection(&g_csDatabase);

	//
	// O[oϐݒ肷
	//
	char modulePath[MAX_PATH * 4];

	GetModuleFileName(g_hInstance, modulePath, sizeof(modulePath));
	char* wp = (char*) _mbsrchr((unsigned char*) modulePath, '\\');
	if (wp) {
		*wp = '\0';
	}

	// BkBsfilWrapper ̃vOtH_
	_snprintf(g_programFolder, sizeof(g_programFolder), "%s\\BkBsfilWrapper", modulePath);
	FORCE_ASCIZ(g_programFolder);

	// bsfilter.exe ւ̃pX
	_snprintf(g_bsfilter, sizeof(g_bsfilter), "%s\\bsfilter.exe", g_programFolder);
	FORCE_ASCIZ(g_bsfilter);

	// iconv.dll ւ̃pX
	char iconv[MAX_PATH * 4];
	_snprintf(iconv, sizeof(iconv), "%s\\iconv.dll", g_programFolder);
	FORCE_ASCIZ(iconv);

	// Becky! ̃f[^tH_
	std::string bkaDataFolder = bka.GetDataFolder();
	if (!Util::EndsWithPathDelimiter(bkaDataFolder)) {
		bkaDataFolder += "\\";
	}
	strncpy(g_bkaDataFolder, bkaDataFolder.c_str(), sizeof(g_bkaDataFolder));
	FORCE_ASCIZ(g_bkaDataFolder);

	// vOC̃f[^tH_
	_snprintf(g_dataFolder, sizeof(g_dataFolder), "%sPlugIns\\BkBsfilWrapper", g_bkaDataFolder);
	FORCE_ASCIZ(g_dataFolder);

	// BkBsfilWrapper ̐ݒt@C
	_snprintf(g_bkBsfilWrapperConfig, sizeof(g_bkBsfilWrapperConfig), "%s\\" CONFIG_FILENAME, g_dataFolder);
	FORCE_ASCIZ(g_bkBsfilWrapperConfig);

	// BkBsfilWrapper ̐ݒt@C(p)
	_snprintf(g_bkBsfilWrapperEnvConfig, sizeof(g_bkBsfilWrapperEnvConfig), "%s\\" CONFIG_ENV_FILENAME, g_dataFolder);
	FORCE_ASCIZ(g_bkBsfilWrapperEnvConfig);

	// bsfilter ̐ݒt@C
	_snprintf(g_bsfilterConfig, sizeof(g_bsfilterConfig), "%s\\" BSFILTER_CONFIG_FILENAME, g_dataFolder);
	FORCE_ASCIZ(g_bsfilterConfig);

	// bsfilter ̃f[^x[XtH_
	_snprintf(g_dbFolder, sizeof(g_dbFolder), "%s\\" BSFILTER_DB_FOLDERNAME, g_dataFolder);
	FORCE_ASCIZ(g_dbFolder);

	// Ot@C̃pX
	_snprintf(g_logFilepath, sizeof(g_logFilepath), "%s\\%s.%s", g_dataFolder, LOG_FILE_NAME, LOG_FILE_EXT);
	FORCE_ASCIZ(g_logFilepath);

	// zCgXg̃pX
	_snprintf(g_whiteListFilepath, sizeof(g_whiteListFilepath), "%s\\" WHITELIST_FILENAME, g_dataFolder);
	FORCE_ASCIZ(g_whiteListFilepath);


	//
	// \[X[h
	//
	CResourceManager::Initialize(g_hInstance, g_programFolder);


	//
	// 쐬
	//

	// vOC̃f[^tH_쐬B
	std::string dataFolder = g_dataFolder;
	dataFolder += "\\";
	rc = MakeSureDirectoryPathExists(dataFolder.c_str());
	if (!rc) {
		return false;
	}

	//
	// ݒǂݍ
	//

	// BkBsfilWrapper ̐ݒǂݍ
	g_bkBsfilProfileList.push_back(&g_bkProfile);
	g_bkBsfilProfileList.push_back(&g_ctxMenuProfile);	

	g_profileManager.Load(g_bkBsfilProfileList, g_bkBsfilWrapperConfig);

	g_bkProfile.SetLockObject(&g_csSetting);
	g_ctxMenuProfile.SetLockObject(&g_csSetting);

	// BkBsfilWrapper ̐ݒ(p)ǂݍ
	g_bkBsfilProfileEnvList.push_back(&g_bkProfileEnv);

	g_profileManagerEnv.Load(g_bkBsfilProfileEnvList, g_bkBsfilWrapperEnvConfig);

	g_bkProfileEnv.SetLockObject(&g_csSetting);

	// Oݒ
	CLogger::Initialize(g_logFilepath, g_dataFolder, g_bkProfile.GetLogLevel().c_str());

	//
	// !!!!! OŃOo͂Ȃ!!!!!
	//

	LOGGING_DEBUG("begin.");


	//
	// g̃[AhX̏擾
	//
	g_myMailList.SetBkDataFolder(g_bkaDataFolder);
	g_myMailList.Update();

	//
	// zCgXgǂݍ
	//
	g_whiteList.Import(g_whiteListFilepath);
	{
		LOGGING_DEBUG("load g_whiteList.");
		std::map<std::string, MailAddressInfo>& whiteMailList = g_whiteList.GetMailAddressList();
		for (std::map<std::string, MailAddressInfo>::iterator it = whiteMailList.begin(); it != whiteMailList.end(); ++it) {
			LOGGING_DEBUG("g_whiteList: [%s].", it->first.c_str());
		}
	}
	g_whiteList.SetLockObject(&g_csSetting);

	//
	// bsfilter ֘A̐ݒsB
	//

	struct _stat fsts;

	// bsfilter.exe ̋N iconv.dll ȂƂ邽߁A
	// PATH  BkBsfilWrapper ̃vOtH_ǉB
	char path[8192];
	char pathnew[8192];

	pcfnname = "GetEnvironmentVariable()";
	rc = GetEnvironmentVariable("PATH", path, sizeof(path) - 1);
	if (!rc) {
		LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
		return false;
	}
	LOGGING_DEBUG("%s: succeeded.", pcfnname);
	FORCE_ASCIZ(path);

	_snprintf(pathnew, sizeof(pathnew), "%s;%s", path, g_programFolder);
	FORCE_ASCIZ(pathnew);
	pcfnname = "SetEnvironmentVariable()";
	rc = SetEnvironmentVariable("PATH", pathnew);
	if (!rc) {
		LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
	}
	LOGGING_INFO("%s: succeeded. path: [%s].", pcfnname, pathnew);

	// bsfilter ̐ݒt@C΁A쐬
	if (_stat(g_bsfilterConfig, &fsts) != 0) {
		CreateBsfilterConfig();
	}

	// bsfilter ̃f[^x[XtH_쐬B
	std::string dbFolder = g_dbFolder;
	dbFolder += "\\";
	pcfnname = "MakeSureDirectoryPathExists()";
	rc = MakeSureDirectoryPathExists(dbFolder.c_str());
	if (!rc) {
		LOGGING_ERROR("%s: LastError: [%lu].", pcfnname, GetLastError());
		return false;
	}
	LOGGING_DEBUG("%s: succeeded.", pcfnname);

	// bsfilter  isconv ݂̑mF
	bool isExistBsfilter = false;
	bool isExistIconv = false;

	if (_stat(g_bsfilter, &fsts) == 0 && fsts.st_mode & _S_IFREG) {
		isExistBsfilter = true;
	} else {
		LOGGING_INFO("bsfilter.exe not exist. g_bsfilter: [%s].", g_bsfilter);
	}
	if (_stat(iconv, &fsts) == 0 && fsts.st_mode & _S_IFREG) {
		isExistIconv = true;
	} else {
		LOGGING_INFO("iconv.dll not exist. iconv: [%s].", iconv);
	}

	if (isExistBsfilter && isExistIconv) {

		//
		// bsfilter  isconv ݂Ƃ݂̂͗pł邩mF
		//

		if (g_bkProfile.GetDontCheckDbOnStart()) {

			// ݒŃf[^x[XQƂȂƂȂĂ̂ŁAmFɗpł̂Ƃ݂ȂB
			LOGGING_INFO("g_bkProfile.DontCheckDbOnStart: [true].");
			g_isUseBsfilter = TRUE;

		} else {

			// bsfilter pł邩mF邽߂Ƀf[^x[XQƂĂ݂B
			std::string output;

			pcfnname = "ExecBsfilter()";
			int status = ExecBsfilter(0, 0, &output, "--show-db-status");
			LOGGING_INFO("%s: status: [%d].", pcfnname, status);

			// bsfilter.exe 𗘗pł邩ǂ
			g_isUseBsfilter = (0 == status) ? TRUE : FALSE;
			LOGGING_INFO("g_isUseBsfilter: [%s].", (g_isUseBsfilter) ? "TRUE" : "FALSE");
		}

	} else {
		LOGGING_INFO("!isExistBsfilter || !isExistIconv.");
	}

	LOGGING_DEBUG("end.");

	return true;

#undef __FUNCNAME__
}

//----------------------------------------------------------------------
// Is
//----------------------------------------------------------------------

void Release()
{
	// \[X
	CResourceManager::Release();

	// ݒ̔rpIuWFNg폜
	DeleteCriticalSection(&g_csSetting);

	// f[^x[X̔rpIuWFNg폜
	DeleteCriticalSection(&g_csDatabase);
}

//----------------------------------------------------------------------
// spam Ɣ肳ꂽ[u
//----------------------------------------------------------------------

int ProcessSpamMail(LPCSTR lpMessage, LPCSTR lpMailBox, int* lpnAction, char** lppParam)
{
	// NOTE: Ô *lppParam ɃS~܂͉ς݃|C^Ă\lAݒ̂ݍsAQƂȂƁB

	int result = BKC_FILTER_DEFAULT;
	*lpnAction = ACTION_NOTHING;

	switch (g_bkProfile.GetProcSpam()) {
	case BkBsfilProfile::GoTrash: {
			const char* trashFolder = "\\!Trash\\";

			char* rcTrashFolder = (char*) bka.Alloc(strlen(trashFolder) + 1);
			strcpy(rcTrashFolder, trashFolder);

			*lpnAction = ACTION_MOVEFOLDER;
			*lppParam = rcTrashFolder;
			result = BKC_FILTER_DONE;
		}
		break;

	case BkBsfilProfile::GoFolder:
		if (0 < g_bkProfile.GetSpamFolder().length()) {

			std::string folderId = g_bkProfile.GetSpamFolder();

			if (g_bkProfile.GetGoFolderSameRelativeFolderPath()) {

				//
				// tH_ֈړꍇɊe[{bNX̓ʒũtH_Ɉړ
				//

				// [{bNX݂̃[{bNXɕύX
				std::string mailBox = lpMailBox;
				folderId = ChangeMailboxId(folderId, mailBox);
			}

			char* spamFolder = (char*) bka.Alloc(folderId.length() + 1);
			strcpy(spamFolder, folderId.c_str());

			*lpnAction = ACTION_MOVEFOLDER;
			*lppParam = spamFolder;
			result = BKC_FILTER_DONE;
		}
		break;

	case BkBsfilProfile::AddHeader:
		if (0 < g_bkProfile.GetSpamHeaderFieldName().length() && 0 < g_bkProfile.GetSpamHeaderFieldBody().length()) {

			std::string addHeader;
			GET_SPAM_HEADER(&addHeader, &g_bkProfile);

			char* rcAddHeader = (char*) bka.Alloc(addHeader.length() + 1);
			strcpy(rcAddHeader, addHeader.c_str());

			*lpnAction = ACTION_ADDHEADER;
			*lppParam = rcAddHeader;
			result = BKC_FILTER_PASS;
		}
		break;

	case BkBsfilProfile::NoMove:
		// Ȃ
		break;
	}

	return result;
}

//----------------------------------------------------------------------
// spam  token f[^x[Xɑ
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_AddSpam(HWND hWnd, LPARAM lParam)
{
	LearnAndUpdate(hWnd, "--add-spam");
}

//----------------------------------------------------------------------
// clean ȃ[ token f[^x[Xɑ
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_AddClean(HWND hWnd, LPARAM lParam)
{
	LearnAndUpdate(hWnd, "--add-clean");
}

//----------------------------------------------------------------------
// spam  token f[^x[X
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_SubSpam(HWND hWnd, LPARAM lParam)
{
	LearnAndUpdate(hWnd, "--sub-spam");
}

//----------------------------------------------------------------------
// clean ȃ[ token f[^x[X
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_SubClean(HWND hWnd, LPARAM lParam)
{
	LearnAndUpdate(hWnd, "--sub-clean");
}

//----------------------------------------------------------------------
// mf[^x[XXV
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_Update(HWND hWnd, LPARAM lParam)
{
	std::string message;
	std::string title;

	CWaitCursor wait;

	std::string output;
	int rc = ExecBsfilter(0, 0, &output, "--update");
	if (rc == 0) {

		// XV̏Ԃ𓾂
		std::string stsoutput;
		int status = ExecBsfilter(0, 0, &stsoutput, "--show-db-status");

		message = CResourceManager::FormatResourceString(IDS_MSG_DATABASE_UPDATED);
		if (status == 0) {
			Util::SameCrLf(stsoutput);
			message += "\r\n\r\n";
			message += stsoutput;
		}
		title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_DATABASE_UPDATE_RESULTS);

		if (!g_bkProfile.GetDisableResultMsgbox()) {
			wait.Restore();
			MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONINFORMATION);
		}
	} else {
		message = CResourceManager::FormatResourceString(IDS_MSG_DATABASE_UPDATE_FAILED);
		title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_DATABASE_UPDATE_RESULTS);

		wait.Restore();
		MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONERROR);
	}
}

//----------------------------------------------------------------------
// f[^x[X̏Ԃ\
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_ShowDbStatus(HWND hWnd, LPARAM lParam)
{
	std::string message;
	std::string title;

	CWaitCursor wait;

	std::string stsoutput;
	int status = ExecBsfilter(0, 0, &stsoutput, "--show-db-status");
	if (status == 0) {

		Util::SameCrLf(stsoutput);
		message = CResourceManager::FormatResourceString(IDS_MSG_ACQUIRED_STATE_OF_DATABASE);
		message += "\r\n\r\n";
		message += stsoutput;
		title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_STATE_OF_DATABASE);

		wait.Restore();
		MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONINFORMATION);
	} else {
		message = CResourceManager::FormatResourceString(IDS_MSG_FAILED_TO_GET_DATABASE_STATE);
		title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_STATE_OF_DATABASE);

		wait.Restore();
		MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONERROR);
	}
}

//----------------------------------------------------------------------
// [ spam Ƃďu
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_ToSpam(HWND hWnd, LPARAM lParam)
{
	std::string message;
	std::string title;
	int rc;

	message = CResourceManager::FormatResourceString(IDS_MSG_BEFORE_EMAIL_SPAM_ACTION);
	title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_BEFORE_EMAIL_SPAM_ACTION);

	if (!g_bkProfile.GetDisableBeforeMsgbox()) {
		rc = MessageBox(hWnd, message.c_str(), title.c_str(), MB_OKCANCEL | MB_ICONINFORMATION);
		if (rc != IDOK) {
			return;
		}
	}

	const char* pCommand = "--add-spam";

	//
	// I𒆂̃[ spam ƂĊwKAmf[^x[XXV
	//
	std::vector<std::string> mailIdList;
	std::vector<std::string> processedMailIdList;

	GetSelectedMailIdList(&mailIdList);

	// i\p
	HWND hWndMain = 0;
	HWND hWndDummy = 0;
	bka.GetWindowHandles(&hWndMain, &hWndDummy, &hWndDummy, &hWndDummy);

	CProgressDlg progressDlg(mailIdList.size(), title.c_str());
	progressDlg.Show(hWndMain);

	int learnCount = LearnMail(&mailIdList, &processedMailIdList, pCommand, progressDlg);

	std::string terminateStepMsg = CResourceManager::FormatResourceString(IDS_MSG_PROGRESS_TERMINATE);
	progressDlg.TerminateStep(terminateStepMsg.c_str());

	int updateStatus = 0;
	if (0 < learnCount) {
		std::string output;

		updateStatus = ExecBsfilter(0, 0, &output, "--update");
	}

	if (learnCount == 0 || (0 < learnCount && updateStatus == 0)) {
		// OK
	} else {
		// s
		message = CResourceManager::FormatResourceString(IDS_MSG_EMAIL_FAILED_TO_LEARN);
		title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_SPAM_ACTION_RESULTS);

		progressDlg.Hide();

		MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONERROR);
		return;
	}

	int numAddHeader = 0;

	std::string spamMailIds;

	for (std::vector<std::string>::iterator it = processedMailIdList.begin(); it != processedMailIdList.end(); ++it) {

		std::string spamMailId = *it;

		if (g_bkProfile.GetSetReadToSpamMail()) {
			//
			// [ǂɂ
			//
			bka.SetStatus(spamMailId.c_str(), MESSAGE_READ, 0);
		}

		if (g_bkProfile.GetProcSpam() == BkBsfilProfile::GoTrash
		 || g_bkProfile.GetProcSpam() == BkBsfilProfile::GoFolder) {

			// [ ID  bka.MoveMessages() ɓn悤AmailId1|mailId2|...|mailIdN ̌`ɂ܂Ƃ߂
			if (spamMailIds.length() != 0) {
				spamMailIds += "|";
			}
			spamMailIds += spamMailId;

		} else if (g_bkProfile.GetProcSpam() == BkBsfilProfile::AddHeader) {

			char* pMail = bka.GetSource(spamMailId.c_str());
			if (pMail && 0 < g_bkProfile.GetSpamHeaderFieldName().length() && 0 < g_bkProfile.GetSpamHeaderFieldBody().length()) {

				//
				// wb_ǉ
				//
				int mailLength = strlen(pMail);
				bool brc = AddSpamHeader(spamMailId, pMail, mailLength);
				if (brc) {
					++numAddHeader;
				}
			}
			if (pMail) {
				bka.Free(pMail);
			}
		}
	}

	title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_SPAM_ACTION_RESULTS);

	if (g_bkProfile.GetProcSpam() == BkBsfilProfile::GoTrash
	 || g_bkProfile.GetProcSpam() == BkBsfilProfile::GoFolder) {

		//
		// [ړ
		//
		std::string folderId = GetSpamFolderId();
		if (0 < folderId.length() && 0 < spamMailIds.length()) {
			bka.MoveMessages(folderId.c_str(), spamMailIds.c_str(), FALSE);
		}

	} else if (g_bkProfile.GetProcSpam() == BkBsfilProfile::AddHeader) {
		
		std::vector<std::string>::iterator it;

		// spam ƂĊwK[̂ݑI
		for (it = mailIdList.begin(); it != mailIdList.end(); ++it) {

			std::vector<std::string>::iterator itProcessed = std::find(processedMailIdList.begin(), processedMailIdList.end(), *it);
			if (itProcessed != processedMailIdList.end()) {
				// spam ƂĊwK[I
				bka.SetSel(it->c_str(), TRUE);
			} else {
				// ȊÕ[̑I
				bka.SetSel(it->c_str(), FALSE);
			}
		}

		// I𒆂̃[U蕪
		bka.Command(NULL, "FilterSelected");
	}

	progressDlg.Hide();

	message = CResourceManager::FormatResourceString(IDS_MSG_LEARNED_EMAIL_TO_SPAM,
			mailIdList.size(), learnCount
		);

	if (!g_bkProfile.GetDisableResultMsgbox()) {
		MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONINFORMATION);
	}
}

//----------------------------------------------------------------------
// [ spam wb_
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_RemoveSpamHeader(HWND hWnd, LPARAM lParam)
{
	std::string message;
	std::string title;
	int rc;

	message = CResourceManager::FormatResourceString(IDS_MSG_BEFORE_REMOVE_SPAM_HEADER);
	title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_BEFORE_REMOVE_SPAM_HEADER);

	if (!g_bkProfile.GetDisableBeforeMsgbox()) {
		rc = MessageBox(hWnd, message.c_str(), title.c_str(), MB_OKCANCEL | MB_ICONINFORMATION);
		if (rc != IDOK) {
			return;
		}
	}


	//
	// I𒆂̃[ spam wb_
	//
	std::vector<std::string> mailIdList;

	GetSelectedMailIdList(&mailIdList);

	// i\p
	HWND hWndMain = 0;
	HWND hWndDummy = 0;
	bka.GetWindowHandles(&hWndMain, &hWndDummy, &hWndDummy, &hWndDummy);

	CProgressDlg progressDlg(mailIdList.size(), title.c_str());
	progressDlg.Show(hWndMain);

	int numRemoveHeader = 0;

	for (std::vector<std::string>::iterator it = mailIdList.begin(); it != mailIdList.end(); ++it) {

		std::string mailId = *it;

		char* pMail = bka.GetSource(mailId.c_str());
		if (!pMail) {
			continue;
		}

		int mailLength = strlen(pMail);

		bool bRemove = RemoveSpamHeader(mailId, pMail, mailLength);
		if (bRemove) {
			++numRemoveHeader;
		}

		bka.Free(pMail);

		// i\s
		progressDlg.StepIt();

		if (progressDlg.IsCancel()) {
			break;
		}
	}

	std::string terminateStepMsg = CResourceManager::FormatResourceString(IDS_MSG_PROGRESS_TERMINATE);
	progressDlg.TerminateStep(terminateStepMsg.c_str());

	progressDlg.Hide();

	message = CResourceManager::FormatResourceString(IDS_MSG_REMOVE_SPAM_HEADER_RESULTS,
			mailIdList.size(), numRemoveHeader
		);
	title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_REMOVE_SPAM_HEADER_RESULTS);

	if (!g_bkProfile.GetDisableResultMsgbox()) {
		MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONINFORMATION);
	}
}

//----------------------------------------------------------------------
// [𔻒肵u
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_TestMailAndFilter(HWND hWnd, LPARAM lParam)
{
	std::string message;
	std::string title;
	int rc;

	message = CResourceManager::FormatResourceString(IDS_MSG_BEFORE_EMAIL_JUDGE_AND_ACTION);
	title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_BEFORE_EMAIL_JUDGE_AND_ACTION);

	if (!g_bkProfile.GetDisableBeforeMsgbox()) {
		rc = MessageBox(hWnd, message.c_str(), title.c_str(), MB_OKCANCEL | MB_ICONINFORMATION);
		if (rc != IDOK) {
			return;
		}
	}


	//
	// I𒆂̃[𔻒肵Au
	//
	std::vector<std::string> mailIdList;

	GetSelectedMailIdList(&mailIdList);

	// i\p
	HWND hWndMain = 0;
	HWND hWndDummy = 0;
	bka.GetWindowHandles(&hWndMain, &hWndDummy, &hWndDummy, &hWndDummy);

	CProgressDlg progressDlg(mailIdList.size(), title.c_str());
	progressDlg.Show(hWndMain);

	int numAddHeader = 0;

	std::vector<std::string> spamMailIdList;
	std::vector<std::string> cleanMailIdList;

	for (std::vector<std::string>::iterator it = mailIdList.begin(); it != mailIdList.end(); ++it) {

		std::string mailId = *it;

		char* pMail = bka.GetSource(mailId.c_str());
		if (!pMail) {
			continue;
		}

		int mailLength = strlen(pMail);
		std::string output;
		int rc = ExecBsfilter(pMail, mailLength, &output, "");
		if (rc == 0) {
			//
			// spam Ɣ肳ꂽ
			//
			spamMailIdList.push_back(mailId);

			if (g_bkProfile.GetSetReadToSpamMail()) {
				// [ǂɂ
				bka.SetStatus(mailId.c_str(), MESSAGE_READ, 0);
			}

			if (g_bkProfile.GetProcSpam() == BkBsfilProfile::AddHeader) {

				//
				// wb_ǉ
				//
				bool brc = AddSpamHeader(mailId, pMail, mailLength);
				if (brc) {
					++numAddHeader;
				}
			}

			if (g_bkProfile.GetRemoveSpamFromWhiteList()) {
				 // 莞у[wK spam ȃ[zCgXg珜
				RemoveMailOrDomainFromWhiteList(pMail);
			}

		} else {
			cleanMailIdList.push_back(mailId);

			if (!g_bkProfile.GetDontRemoveSpamHeader()) {
				// 莞у[wK clean ȃ[ spam wb_
				RemoveSpamHeader(mailId, pMail, mailLength);
			}
		}

		bka.Free(pMail);

		// i\s
		progressDlg.StepIt();

		if (progressDlg.IsCancel()) {
			break;
		}
	}

	std::string terminateStepMsg = CResourceManager::FormatResourceString(IDS_MSG_PROGRESS_TERMINATE);
	progressDlg.TerminateStep(terminateStepMsg.c_str());

	title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_JUDGE_AND_ACTION_RESULTS);

	if (g_bkProfile.GetProcSpam() == BkBsfilProfile::GoTrash
		|| g_bkProfile.GetProcSpam() == BkBsfilProfile::GoFolder) {

		//
		// [ړ
		//
		std::string folderId = GetSpamFolderId();

		std::vector<std::string>::iterator it;

		// spam Ɣ肳ꂽ[̑IA
		// bka.MoveMessages() ɓn悤AmailId1|mailId2|...|mailIdN ̌`ɂ܂Ƃ߂
		std::string spamMailIds;
		for (it = spamMailIdList.begin(); it != spamMailIdList.end(); ++it) {

			bka.SetSel(it->c_str(), FALSE);

			if (spamMailIds.length() != 0) {
				spamMailIds += "|";
			}
			spamMailIds += *it;
		}

		// ړ
		if (0 < folderId.length() && 0 < spamMailIds.length()) {
			bka.MoveMessages(folderId.c_str(), spamMailIds.c_str(), FALSE);
		}
	}

	//
	// I
	//
	if (g_bkProfile.GetProcSpam() == BkBsfilProfile::GoTrash
		|| g_bkProfile.GetProcSpam() == BkBsfilProfile::GoFolder
		|| g_bkProfile.GetProcSpam() == BkBsfilProfile::NoMove) {

		// clean Ɣ肳ꂽ[̂ݑI
		for (std::vector<std::string>::iterator it = mailIdList.begin(); it != mailIdList.end(); ++it) {

			std::vector<std::string>::iterator itClean = std::find(cleanMailIdList.begin(), cleanMailIdList.end(), *it);
			if (itClean != cleanMailIdList.end()) {
				// clean Ɣ肳ꂽ[I
				bka.SetSel(it->c_str(), TRUE);
			} else {
				// ȊÕ[̑I
				bka.SetSel(it->c_str(), FALSE);
			}
		}

	} else if (g_bkProfile.GetProcSpam() == BkBsfilProfile::AddHeader) {
	
		// 肵[̂ݑI
		for (std::vector<std::string>::iterator it = mailIdList.begin(); it != mailIdList.end(); ++it) {

			std::vector<std::string>::iterator itClean = std::find(cleanMailIdList.begin(), cleanMailIdList.end(), *it);
			std::vector<std::string>::iterator itSpam = std::find(spamMailIdList.begin(), spamMailIdList.end(), *it);

			if (itClean != cleanMailIdList.end()) {
				// clean Ɣ肳ꂽ[I
				bka.SetSel(it->c_str(), TRUE);
			} else if (itSpam != spamMailIdList.end()) {
				// spam Ɣ肳ꂽ[I
				bka.SetSel(it->c_str(), TRUE);
			} else {
				// ȊÕ[̑I
				bka.SetSel(it->c_str(), FALSE);
			}
		}
	}


	// I𒆂̃[U蕪
	bka.Command(NULL, "FilterSelected");

	progressDlg.Hide();

	message = CResourceManager::FormatResourceString(IDS_MSG_JUDGED_EMAIL_TO_SPAM,
			mailIdList.size(), spamMailIdList.size()
		);

	if (!g_bkProfile.GetDisableResultMsgbox()) {
		MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONINFORMATION);
	}
}

//----------------------------------------------------------------------
// [𔻒(eXg)
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_TestMail(HWND hWnd, LPARAM lParam)
{
	std::string message;
	std::string title;

	CWaitCursor wait;

	char* pMail = bka.GetSource(0);
	if (!pMail) {
		return;
	}

	int length = strlen(pMail);
	std::string output;
	int rc = ExecBsfilter(pMail, length, &output, "");
	bka.Free(pMail);

	Util::SameCrLf(output);

	if (rc == 0) {
		message = CResourceManager::FormatResourceString(IDS_MSG_JUDGE_TEST_MAIL_TO_SPAM);

	} else if (rc == 1) {
		message = CResourceManager::FormatResourceString(IDS_MSG_JUDGE_TEST_MAIL_TO_CLEAN);

	} else {
		message = CResourceManager::FormatResourceString(IDS_MSG_JUDGE_TEST_FAILED);
	}
	message += "\r\n\r\n";
	message += output;

	title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_JUDGE_TEST_RESULTS);

	wait.Restore();
	MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONINFORMATION);
}

//----------------------------------------------------------------------
// zCgXgւ̒ǉE폜s
//----------------------------------------------------------------------

typedef int (*FP_ADD_OR_UPDATE_WHITELIST_CALLBACK)(const char* name, const char* email);

static void UpdateWhiteList(HWND hWnd, UINT idBeforeMsg, UINT idBeforeCaption, UINT idResultMsg, UINT idResultCaption, FP_ADD_OR_UPDATE_WHITELIST_CALLBACK callbackf)
{
	std::string message;
	std::string title;
	int rc;

	message = CResourceManager::FormatResourceString(idBeforeMsg);
	title = CResourceManager::FormatResourceString(idBeforeCaption);

	if (!g_bkProfile.GetDisableBeforeMsgbox()) {

		rc = MessageBox(hWnd, message.c_str(), title.c_str(), MB_OKCANCEL | MB_ICONINFORMATION);
		if (rc != IDOK) {
			return;
		}
	}


	std::vector<std::string> mailIdList;

	GetSelectedMailIdList(&mailIdList);

	// i\p
	HWND hWndMain = 0;
	HWND hWndDummy = 0;
	bka.GetWindowHandles(&hWndMain, &hWndDummy, &hWndDummy, &hWndDummy);

	CProgressDlg progressDlg(mailIdList.size(), title.c_str());
	progressDlg.Show(hWndMain);

	int effectCount = 0;

	for (std::vector<std::string>::iterator it = mailIdList.begin(); it != mailIdList.end(); ++it) {

		char* pMail = bka.GetSource(it->c_str());
		if (!pMail) {
			continue;
		}

		char name[1024];
		char email[512];

		std::string from = Util::GetHeaderFieldBody(pMail, "From");
		GetDecodedNameAndAddr(name, sizeof(name), email, sizeof(email), from.c_str());

		// s
		int count = callbackf(name, email);
		effectCount += count;


		bka.Free(pMail);

		// i\s
		progressDlg.StepIt();

		if (progressDlg.IsCancel()) {
			break;
		}
	}

	g_whiteList.Export(g_whiteListFilepath);

	std::string terminateStepMsg = CResourceManager::FormatResourceString(IDS_MSG_PROGRESS_TERMINATE);
	progressDlg.TerminateStep(terminateStepMsg.c_str());

	progressDlg.Hide();

	message = CResourceManager::FormatResourceString(idResultMsg,
			effectCount
		);
	title = CResourceManager::FormatResourceString(idResultCaption);

	if (!g_bkProfile.GetDisableResultMsgbox()) {
		MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONINFORMATION);
	}
}

//----------------------------------------------------------------------
// olzCgXgɒǉ
//----------------------------------------------------------------------

static int AddToWhiteList(const char* name, const char* email)
{
	bool bNoteIsEmpty = false;
	if (strlen(email) != 0 && !g_myMailList.IsExist(email) && g_whiteList.IsNotExistOrNoteIsEmpty(email, &bNoteIsEmpty)) {
		// g̃[AhX͓o^Ȃ
		// ݂ȂAo^ς݂̔lŁAAo^悤ƂAhX̔lłȂꍇ̂݁Ao^
		std::string note = Util::Trim(name);
		if (!bNoteIsEmpty || (bNoteIsEmpty && 0 < note.length())) {
			g_whiteList.AddOrUpdate(email, note.c_str());
			return 1;
		}
	}

	return 0;
}

void WINAPI Menu_CmdProc_AddToWhiteList(HWND hWnd, LPARAM lParam)
{
	UpdateWhiteList(
			hWnd,
			IDS_MSG_BEFORE_ADD_SENDER_TO_WHITELIST, IDS_MSG_CAPTION_BEFORE_ADD_SENDER_TO_WHITELIST,
			IDS_MSG_ADD_SENDER_TO_WHITELIST, IDS_MSG_CAPTION_ADD_SENDER_TO_WHITELIST_RESULTS,
			AddToWhiteList
		);
}

//----------------------------------------------------------------------
// olhCzCgXgɒǉ
//----------------------------------------------------------------------

static int AddDomainToWhiteList(const char* name, const char* email)
{
	std::string domain = Util::GetAtNext(email);

	bool bNoteIsEmpty = false;
	if (domain.length() != 0 && !g_myMailList.IsDomainExist(domain.c_str()) && g_whiteList.IsNotExistOrNoteIsEmpty(domain.c_str(), &bNoteIsEmpty)) {
		// g̃[AhX̃hC͓o^Ȃ
		// ݂ȂAo^ς݂̔lŁAAo^悤ƂAhX̔lłȂꍇ̂݁Ao^
		std::string note = Util::Trim(name);
		if (!bNoteIsEmpty || (bNoteIsEmpty && 0 < note.length())) {
			g_whiteList.AddOrUpdate(domain.c_str(), note.c_str());
			return 1;
		}
	}

	return 0;
}

void WINAPI Menu_CmdProc_AddDomainToWhiteList(HWND hWnd, LPARAM lParam)
{
	UpdateWhiteList(
			hWnd,
			IDS_MSG_BEFORE_ADD_SENDER_DOMAIN_TO_WHITELIST, IDS_MSG_CAPTION_BEFORE_ADD_SENDER_DOMAIN_TO_WHITELIST,
			IDS_MSG_ADD_SENDER_DOMAIN_TO_WHITELIST, IDS_MSG_CAPTION_ADD_SENDER_DOMAIN_TO_WHITELIST_RESULTS,
			AddDomainToWhiteList
		);
}

//----------------------------------------------------------------------
// olzCgXg폜
//----------------------------------------------------------------------

static int RemoveFromWhiteList(const char* name, const char* email)
{
	if (strlen(email) != 0 && g_whiteList.IsExist(email)) {
		g_whiteList.Delete(email);
		return 1;
	}

	return 0;
}

void WINAPI Menu_CmdProc_RemoveFromWhiteList(HWND hWnd, LPARAM lParam)
{
	UpdateWhiteList(
			hWnd,
			IDS_MSG_BEFORE_REMOVE_SENDER_FROM_WHITELIST, IDS_MSG_CAPTION_BEFORE_REMOVE_SENDER_FROM_WHITELIST,
			IDS_MSG_REMOVE_SENDER_FROM_WHITELIST, IDS_MSG_CAPTION_REMOVE_SENDER_FROM_WHITELIST_RESULTS,
			RemoveFromWhiteList
		);
}

//----------------------------------------------------------------------
// olhCzCgXg폜
//----------------------------------------------------------------------

static int RemoveDomainFromWhiteList(const char* name, const char* email)
{
	std::string domain = Util::GetAtNext(email);

	if (domain.length() != 0 && g_whiteList.IsExist(domain.c_str())) {
		g_whiteList.Delete(domain.c_str());
		return 1;
	}

	return 0;
}

void WINAPI Menu_CmdProc_RemoveDomainFromWhiteList(HWND hWnd, LPARAM lParam)
{
	UpdateWhiteList(
			hWnd,
			IDS_MSG_BEFORE_REMOVE_SENDER_DOMAIN_FROM_WHITELIST, IDS_MSG_CAPTION_BEFORE_REMOVE_SENDER_DOMAIN_FROM_WHITELIST,
			IDS_MSG_REMOVE_SENDER_DOMAIN_FROM_WHITELIST, IDS_MSG_CAPTION_REMOVE_SENDER_DOMAIN_FROM_WHITELIST_RESULTS,
			RemoveDomainFromWhiteList
		);
}

//----------------------------------------------------------------------
// bsfilter.exe ̔zu𑣂bZ[W\
//----------------------------------------------------------------------

void WINAPI Menu_CmdProc_ShowPleaseInstallBsfilter(HWND hWnd, LPARAM lParam)
{
	std::string message;
	std::string title;

	message = CResourceManager::FormatResourceString(IDS_MSG_BSFILTER_UNAVAILABLE,
			g_programFolder
		);
	title = CResourceManager::FormatResourceString(IDS_MSG_CAPTION_BSFILTER_UNAVAILABLE);

	MessageBox(hWnd, message.c_str(), title.c_str(), MB_OK | MB_ICONERROR);
}

//----------------------------------------------------------------------
// [̑IԂɂčڂ̕\E\ݒ肷
//----------------------------------------------------------------------

UINT WINAPI Menu_CmdUIProc(HWND hWnd, LPARAM lParam)
{
	// IΏۂ̃[ꍇ̓j[ڂIłȂ悤ɂ
	// NOTE: Ɨǂ@΁ABBB
	UINT result = 0;

	if (!g_isUseBsfilter) {
		result |= BKMENU_CMDUI_DISABLED;
		return result;
	}

	char mailId[256];

	int itemPosition = bka.GetNextMail(-1, mailId, sizeof(mailId), TRUE);
	if (itemPosition == -1) {
		result |= BKMENU_CMDUI_DISABLED;
	}

	return result;
}

